#!/usr/local/bin/perl

use strict;

#==================================================================
# WwwLounge 3.0
#
# Copyright (C) 1997-2021 杜甫々 https://www.tohoho-web.com/
#
# 法人/個人/私用/商用を問わず、使用・改造・再配布可能です。
#==================================================================

# ★ラウンジソフト名
# 空文字列("")にしても別の名称に変更しても構いません。
my $g_name = "WwwLounge 3.0";

# ★メール送信コマンド名
# 設置するサーバの仕様に合わせて変更してください。
# メール送信機能を利用しない場合は空文字("")を指定してください。
my $g_sendmail = '/usr/sbin/sendmail';

# ★メールの差出人アドレス
# メール送信機能を利用する場合、メールのFromアドレスを指定してください。
# サーバの設定によっては、本人以外のメールアドレスを指定できなかったり、
# 指定しても自動的に変更されたりします。
my $g_from_addr = 'WwwLounge Admin <admin@example.com>';

# ★管理者パスワード
# 管理者機能(deleteとutime)を使用する場合は管理者パスワードを指定してください。
# 管理者機能を使用しない場合は不要です。
my $g_passwd = "";

# ★戻るリンク
# トップページに[戻る]リンクを表示する場合そのリンク先を設定してください。
# 表示しない場合は空文字("")を設定してください。
my $g_return_url = "../index.html";

# ★使用上の注意リンク
# [使用上の注意]リンクを表示する場合はそのリンク先を設定してください。
# 表示しない場合は空文字("")を設定してください。
my $g_note_url = "note.html";

# ★質問者表示
# トピック一覧画面で質問者を表示する場合は1を、表示しない場合は0を
# 設定してください。
my $g_show_questioner = 1;

# ★トピック一覧最大表示件数
# トピック一覧に表示する最大件数を指定してください。
my $g_topic_list_max_count = 50;

# ★データディレクトリ
# データを保存するディレクトリ名を指定してください。
my $g_dir = "wwwlng";

# ★検索ログファイル名
# 検索ログを記録する場合はそのファイル名を指定してください。
# データディレクトリからの相対パスで指定してください。
# 記録しない場合は空文字("")を指定してください。
my $g_search_log = "../search.log";

# ★ロックディレクトリとロックファイル名
# ロックファイルを作成するディレクトリ名とファイル名を指定してください。
# データディレクトリからの相対パスで指定します。
my $g_lockdir = "../lock";
my $g_lockfile = "$g_lockdir/wwwlng.loc";

# ★タイムゾーン設定
# 通常は日本時刻("JST-9")で構いません。
$ENV{'TZ'} = "JST-9";

# グローバル変数
my @g_wdays = ( "日", "月", "火", "水", "木", "金", "土" );
my %FORM;

#
# メインルーチン
#
{
	# カレントディレクトリを移動する
	chdir($g_dir);

	# フォームデータを読み込む
	if (readForm()) {
		showErrorPage("データの読み込みに失敗しました。");
		exit(0);
	}

	# それぞれのページ表示にジャンプ
	if ($#ARGV == -1) {
		showTopPage();						# トップページ
	} elsif ($ARGV[0] eq "dir") {
		showDirPage($ARGV[1]);				# ディレクトリ指定一覧表示
	} elsif ($ARGV[0] eq "list") {
		showTopicListByDays($ARGV[1]);		# 過去N日分一覧表示
	} elsif ($ARGV[0] eq "print") {
		showTopicPage($ARGV[1]);			# トピック表示
	} elsif ($ARGV[0] eq "showall") {
		showLatelyPage($ARGV[1]);			# 過去N日分全文表示
	} elsif ($ARGV[0] eq "search") {
		showSearchPage();					# 検索表示
	} elsif ($ARGV[0] eq "new") {
		showNewTopicPage();					# 新規トピック入力
	} elsif ($ARGV[0] eq "new2") {
		appendTopic();						# トピック追加
	} elsif ($ARGV[0] eq "append") {
		appendMessage($ARGV[1], 1);			# 発言追加
	} elsif ($ARGV[0] eq "test") {
		showTestPage();						# 自己診断
	} elsif ($ARGV[0] eq "delete") {
		shift(@ARGV);
		deleteTopic(@ARGV);					# トピック削除
	} elsif ($ARGV[0] eq "utime") {
		shift(@ARGV);
		utimeTopic(@ARGV);					# ファイルの更新時刻を合わせる
	} else {
		showErrorPage("URLの指定が誤っています。");
	}

	exit(0);
}

#
# トップページ
#
sub showTopPage {
	my($dir, @dirs, $file, @files, $mtime, $timestr);
	local(*DIR, *DIR2);

	printHeader("ラウンジ(最近更新されたトピック)");
	print "<div class='header_links'>\n";
	if ($g_return_url) {
		print " <a href='$g_return_url'>[戻る]</a>\n";
	}
	if ($g_note_url) {
		print " <a href='$g_note_url'>[使用上の注意]</a>\n";
	}
	print " <a href='wwwlng.cgi?new'>[新規質問]</a>\n";
	print "</div>\n";
	print "<hr>\n";

	# 最近更新されたトピックの一覧を得る
	if (!opendir(DIR, ".")) {
		showErrorPage("フォルダ情報を読み出せません。");
		exit(0);
	}
	while ($dir = readdir(DIR)) {
		if ($dir =~ /^\./) { next; }
		if (! -d $dir) { next; }
		push(@dirs, $dir);
		opendir(DIR2, $dir) || next;
		while ($file = readdir(DIR2)) {
			if ($file =~ /^\./) { next; }
			if (! -f "$dir/$file") { next; }
			$mtime = (stat("$dir/$file"))[9];
			$timestr = formatDateTime("YYYYMMDDhhmmss", $mtime);
			push(@files, "$timestr $dir/$file");
		}
		closedir(DIR2);
	}
	closedir(DIR);
	@dirs = sort(@dirs);
	@files = sort { $b cmp $a } @files;
	splice(@files, $g_topic_list_max_count);

	# 最近更新されたトピック一覧
	printTopicList(@files);

	# フォルダ一覧
	if ($#dirs >= 0) {
		print "<h2><a name='folders'>過去フォルダ</h2>\n";
		print "<ul class='folder_list'>\n";
		foreach $dir (@dirs) {
			print " <li><a href='wwwlng.cgi?dir+$dir'>[$dir]</a>\n";
		}
		print "</ul>\n";
	}

	# フッタ
	printFooter();
}

#
# ディレクトリ指定一覧表示
#
sub showDirPage {
	my($dir) = @_;
	my($file, @files);

	if (($dir !~ /^[\w\/\.]+$/) || ($dir =~ /\.\./) || ($dir =~ /^\//)) {
		showErrorPage("他のフォルダを参照することはできません。");
		exit(0);
	}

	printHeader("ラウンジ(フォルダ$dirのトピック一覧)");

	print "<div class='header_links'>\n";
	print " <a href='wwwlng.cgi'>[戻る]</a>\n";
	print "</div>\n";

	# ファイルの日付とファイル名の一覧を得る
	if (!opendir(DIR, $dir)) {
		showErrorPage("フォルダ情報をえることができませんでした。");
		exit(0);
	}
	while ($file = readdir(DIR)) {
		if ($file =~ /^\./) { next; }
		push(@files, (stat("$dir/$file"))[9] . " $dir/$file");
	}
	closedir(DIR);

	# フォルダのファイル一覧
	printTopicList(@files);

	# フッタを表示する
	printFooter();
}

#
# 過去N日分一覧表示
#
sub showTopicListByDays {
	my($days) = @_;
	my($limittime, $mtime, $timestr);
	my($dir, @dirs, $file, @files);
	local(*DIR, *DIR2);

	printHeader("ラウンジ(過去$days日分一覧表示)");
	print "<div class='header_links'>\n";
	print " <a href='wwwlng.cgi'>[戻る]</a>\n";
	print "</div>\n";

	# 過去n日分のトピック一覧を得る
	$limittime = time() - ($days * 3600 * 24);
	if (!opendir(DIR, ".")) {
		showErrorPage("フォルダ情報の取得に失敗しました。");
		exit(0);
	}
	while ($dir = readdir(DIR)) {
		if ($dir =~ /^\./) { next; }
		if (! -d $dir) { next; }
		push(@dirs, $dir);
		opendir(DIR2, $dir) || next;
		while ($file = readdir(DIR2)) {
			if ($file =~ /^\./) { next; }
			if (! -f "$dir/$file") { next; }
			$mtime = (stat("$dir/$file"))[9];
			if ($mtime >= $limittime) {
				$timestr = formatDateTime("YYYYMMDDhhmmss", $mtime);
				push(@files, "$timestr $dir/$file");
			}
		}
		closedir(DIR2);
	}
	closedir(DIR);
	@files = sort { $b <=> $a } @files;

	# トピックの中身を表示する
	printTopicList(@files);

	printFooter();
}

#
# トピック表示ページ
#
sub showTopicPage {
	my($file) = @_;
	my($subject);
	local(*IN);

	if (($file !~ /^[\w\/\.]+$/) || ($file =~ /\.\./) || ($file =~ /^\//)) {
		showErrorPage("ファイル名が不正です。");
		exit(0);
	}

	if (!open(IN, $file)) {
		showErrorPage("ファイル($file)を読み取れません。");
		exit(0);
	}
	$subject = <IN>;
	$subject =~ s/^Subject: //;
	$subject =~ s/[\r\n]+$//;
	printHeader($subject);
	print "<div class='header_links'>\n";
	print " <a href='wwwlng.cgi'>[戻る]</a>\n";
	print "</div>\n";
	printTopicContent(*IN);
	close(IN);

	print "<h2>追加発言</h2>\n";
	print "<form method='POST' action='wwwlng.cgi?append+$file'>\n";
	print " <div class='append_message'>\n";
	print "  <div class='name'><input type='text' size=50 name='name' placeholder='名前'></div>\n";
	print "  <div class='message'><textarea rows=5 cols=60 name='message' placeholder='追加発言'></textarea></div>\n";
	print "  <div class='solved'><input type='checkbox' name='solved'> ←解決時は質問者本人がここをチェックしてください。</div>\n";
	print " </div>\n";
	print " <button type='submit' class='form_button'>送信</button>\n";
	print "</form>\n";
	printFooter();
}

#
# 過去N日分全文表示
#
sub showLatelyPage {
	my($days) = @_;
	my($limittime, $mtime, $timestr);
	my($dir, @dirs, $file, @files);
	my($subject);
	local(*DIR, *DIR2, *IN);

	printHeader("ラウンジ(全文表示)");
	print "<div class='header_links'>\n";
	print " <a href='wwwlng.cgi'>[戻る]</a>\n";
	print "</div>\n";

	# 過去n日分のトピック一覧を得る
	$limittime = time() - ($days * 3600 * 24);
	if (!opendir(DIR, ".")) {
		showErrorPage("フォルダ情報の取得に失敗しました。");
		exit(0);
	}
	while ($dir = readdir(DIR)) {
		if ($dir =~ /^\./) { next; }
		if (! -d $dir) { next; }
		push(@dirs, $dir);
		opendir(DIR2, $dir) || next;
		while ($file = readdir(DIR2)) {
			if ($file =~ /^\./) { next; }
			if (! -f "$dir/$file") { next; }
			$mtime = (stat("$dir/$file"))[9];
			if ($mtime >= $limittime) {
				$timestr = formatDateTime("YYYYMMDDhhmmss", $mtime);
				push(@files, "$timestr $dir/$file");
			}
		}
		closedir(DIR2);
	}
	closedir(DIR);
	@files = sort { $b <=> $a } @files;

	# トピックの中身を表示する
	foreach $file (@files) {
		$file = (split(/ /, $file))[1];
		if (!open(IN, $file)) {
			next;
		}
		$subject = <IN>;
		$subject =~ s/^Subject: //;
		$subject =~ s/[\r\n]+$//;
		print "<h2>$subject</h2>\n";
		printTopicContent(*IN);
		print "<div class='block'>\n";
		print " <a href='wwwlng.cgi?print+$file'>[この(↑)トピックに追加発言]</a>\n";
		print "</div>\n";
		close(IN);
	}

	printFooter();
}

#
# 新規トピック入力ページ
#
sub showNewTopicPage
{
	printHeader("ラウンジ(新規質問)");

	print "<div class='header_links'>\n";
	print " <a href='wwwlng.cgi'>[戻る]</a>\n";
	print "</div>\n";
	print "<hr>\n";
	print "<form method='POST' action='wwwlng.cgi?new2'>\n";

	print " <div class='form_item'>\n";
	print "  <div class='form_label'>タイトル</div>\n";
	print "  <ul class='form_explain'>\n";
	print "   <li>「～するには？」のような質問形式にしてください。\n";
	print "   <li>「～について」の形式は好ましくありません。\n";
	print "  </ul>\n";
	print "  <input type='text' name='subject'>\n";
	print " </div>\n";

	print " <div class='form_item'>\n";
	print "  <div class='form_label'>名前</div>\n";
	print "  <ul class='form_explain'>\n";
	print "   <li>ペンネームでもかまいません。\n";
	print "  </ul>\n";
	print "  <input type='text' name='name'>\n";
	print " </div>\n";

	print " <div class='form_item'>\n";
	print "  <div class='form_label'>メールアドレス</div>\n";
	print "  <ul class='form_explain'>\n";
	print "   <li>例：foo\@www.xxx.zzz\n";
	print "   <li>半角文字で入力してください。\n";
	print "   <li>空欄でもかまいません。\n";
	print "  </ul>\n";
	print "  <div class='block'>\n";
	print "   <input type='text' name='email'>\n";
	print "  </div>\n";
	if (-x $g_sendmail) {
		print "  <div class='block'>\n";
		print "    <input type='checkbox' name='reply' value='on'> 回答をメールで受信する\n";
		# print "    回答をメールで受信しますか？\n";
		# print "    <input type='radio' name='reply' value='on' CHECKED>はい\n";
		# print "    <input type='radio' name='reply' VALUE=''>いいえ\n";
		print "  </div>\n";
	}
	print " </div>\n";

	print " <div class='form_item'>\n";
	print "  <div class='form_label'>質問内容</div>\n";
	print "  <ul class='form_explain'>\n";
	print "   <li>あまり長文にならないようにしてください。\n";
	print "   <li>エラー発生時、再送信すると二重送信になることがあります。\n";
	print "   <li>回答が得られたら、お礼書き込み時に[解決]チェックしてください。\n";
	print "  </ul>\n";
	print "  <textarea name='message'></textarea>\n";
	print " </div>\n";

	print " <button class='form_button'>送信</button>\n";
	print "</form>\n";
	printFooter();
}

#
# 検索ページ
#
sub showSearchPage
{
	my(@words, $i, $dir, @dirs, $file);
	local(*DIR);

	# 出力時にバッファリングしないようにする
	$| = 1;

	printHeader("ラウンジ(検索)");
	print "<div class='header_links'>\n";
	print " <a href='wwwlng.cgi'>[戻る]</a>\n";
	print "</div>\n";

	# キーワードが無ければ終了
	if (!$FORM{'word'}) {
		goto done;
	}

	# 複数のワードを分割する
	@words = split(/ +/, $FORM{'word'});

	# メタ文字を無効化する
	for ($i = 0; $i <= $#words; $i++) {
		$words[$i] =~ s/([\+\*\.\?\^\$\[\-\]\|\(\)\\])/\\$1/g;
	}

	# 検索キーワードをロギングする
	if (-w $g_search_log) {
		if (open(OUT, ">> $g_search_log")) {
			print OUT formatDateTime("YYYY/MM/DD(A) hh:mm:ss", time());
			print OUT " ";
			for ($i = 0; $i <= $#words; $i++) {
				print OUT "$words[$i] ";
			}
			print OUT "\n";
			close(OUT);
		}
	}

	# ディレクトリの一覧を得る
	@dirs = ();
	if (!opendir(DIR, ".")) {
		showErrorPage("フォルダ情報の取得に失敗しました。");
		exit(0);
	}
	while ($dir = readdir(DIR)) {
		if ($dir =~ /^\./) { next; }
		if (! -d $dir) { next; }
		push(@dirs, $dir);
	}
	closedir(DIR);

	print "<dl class='search_result'>\n";
	foreach $dir (@dirs) {
		if (!opendir(DIR, $dir)) {
			next;
		}
		while ($file = readdir(DIR)) {
			if ($file =~ /^\./) { next; }
			if (! -f "$dir/$file") { next; }
			printSearchResult("$dir/$file", @words);
		}
		closedir(DIR);
	}
	print "</dl>\n";

  done:
	printFooter();
}

#
# 新規トピック追加
#
sub appendTopic
{
	my($file);
	local(*OUT);

	# 入力パラメータのチェック
	if ($FORM{'subject'} =~ /^[ 　\r\n]*$/) {
		showErrorPage("タイトルが入力されていません。");
		exit(0);
	}
	if ($FORM{'subject'} =~ /[\r\n]/) {		# セキュリティ対応
		showErrorPage("タイトルに改行を含めることはできません。");
		exit(0);
	}
	if ($FORM{'name'} =~ /^[ 　\r\n]*$/) {
		showErrorPage("名前が入力されていません。");
		exit(0);
	}
	if ($FORM{'message'} =~ /^[ 　\r\n]*$/) {
		showErrorPage("内容が入力されていません。");
		exit(0);
	}
	if ($FORM{'email'} && ($FORM{'email'} !~ /^[\.\-\w]+@[\.\-\w]+$/)) {
		showErrorPage("メールアドレスの形式が不正です。");
		exit(0);
	}

	# ロックをかける
	if (!fileLock()) {
		showErrorPage("ファイルのロックに失敗しました。");
		exit(0);
	}

	# 使用されていないファイル名を探す
	if (!($file = determineFileName())) {
		showErrorPage("新規質問の書き込みに失敗しました。");
		goto done;
	}

	# サブジェクトを書き込む
	if (!open(OUT, "> $file")) {
		showErrorPage("新規質問の書き込みに失敗しました。");
		goto done;
	}
	print OUT "Subject: $FORM{'subject'}\n";
	if (($FORM{'reply'} eq "on") && $FORM{'email'}) {
		print OUT "Reply-To: $FORM{'email'}\n";
	}
	close(OUT);

	# メッセージを追加する
	appendMessage($file, 0);

  done:
	fileUnlock();
}

#
# 発言追加
#
sub appendMessage {
	my($file, $lockflag) = @_;
	my($date, $msg, $text, $mailto, $old);
	local(*IN, *OUT);

	if (($file !~ /^[\w\/\.]+$/) || ($file =~ /\.\./) || ($file =~ /^\//)) {
		showErrorPage("ファイル名が不正です。");
		exit(0);
	}

	# 入力パラメータのチェック
	if ($FORM{'name'} =~ /^[ 　\r\n]*$/) {
		showErrorPage("名前が入力されていません。");
		exit(0);
	}
	if ($FORM{'message'} =~ /^[ 　\r\n]*$/) {
		showErrorPage("内容が入力されていません。");
		exit(0);
	}

	# ロックをかける
	if ($lockflag) {
		if (!fileLock()) {
			showErrorPage("ファイルのロックに失敗しました。");
			exit(0);
		}
	}

	# トピックに発言を追加する
	$msg = "========================================\n";

	# 名前
	$msg .= "From: $FORM{'name'}\n";
	if ($FORM{'email'} ne "") {
		$FORM{'email'} =~ s/'//g;
		$msg .= "E-Mail: $FORM{'email'}\n";
	}

	# 日付
	$date = formatDateTime("YYYY/MM/DD(A) hh:mm:ss", time());
	$msg .= "Date: $date\n";

	# 解決マーク
	if ($FORM{'solved'} eq "on") {
		$msg .= "Solved: $FORM{'solved'}\n";
	}

	# 空行
	$msg .= "\n";

	# メッセージ
	$text = " " . $FORM{'message'};
	$text =~ s/\r\n$//;
	$text =~ s/\n/\n /mg;
	$msg .= $text;
	$msg .= "\n";

	# ファイルに追記する
	if (!open(OUT, ">> $file")) {
		showErrorPage("$fileへの書きこみに失敗しました。");
		exit(0);
	}
	$old = select(OUT); $| = 1; select($old);
	print OUT $msg;
	close(OUT);

	# 質問者にメールで知らせる
	$mailto = "";
	open(IN, "$file");
	while (<IN>) {
		if (/^Reply-To:/) {
			chop($_);
			$mailto = $_;
			$mailto =~ s/^Reply-To: //;
		}
		if (/^\t/) {
			last;
		}
	}
	if ($mailto ne "") {
		seek(IN, 0, 0);

		open(OUT, "| $g_sendmail -t -i");
		print OUT "To: $mailto\n";
		if ($g_from_addr) {
			print OUT "From: $g_from_addr\n";
		}
		print OUT "Subject: WwwLounge Mail Report\n";
		print OUT "Content-Transfer-Encoding: 8bit\n";
		print OUT "Content-Type: text/plain; charset=\"utf-8\"\n";
		print OUT "\n";
		while (<IN>) {
			$_ =~ s/^ //;
			print OUT $_;
		}
		close(OUT);
	}
	close(IN);

	# ロックをはずす
	if ($lockflag) {
		fileUnlock();
	}

  done:
	# ページを表示する
	print "Location: http://$ENV{'SERVER_NAME'}$ENV{'SCRIPT_NAME'}?print+$file\n\n";

	return(1);
}

#
# 自己診断ページ
#
sub showTestPage
{
	printHeader("ラウンジ(自己診断)");
	print "<div class='header_links'>\n";
	print " <a href='wwwlng.cgi'>[戻る]</a>\n";
	print "</div>\n";
	print "<p>OK. CGIは正常に動作しています。</p>\n";
	if (!-x $g_sendmail) {
		print "<p>NG. $g_sendmailがありません。</p>\n";
	}
	if (!-d $g_lockdir) {
		print "<p>NG. $g_lockdirがありません。</p>\n";
	}
	if (-d $g_lockfile) {
		print "<p>NG. $g_lockfileが残っています。削除してください。</p>\n";
	}
	printFooter();
}

#
# トピックを削除する
#
sub deleteTopic {
	my($passwd, $file) = @_;

	# パスワードをチェックする
	if (($g_passwd eq "") || ($passwd ne $g_passwd)) {
		showErrorPage("パスワードが違います。");
		exit(0);
	}

	if ($file =~ /(^\/|\.\.)/) {
		showErrorPage("他のフォルダのファイルを削除することはできません。");
		exit(0);
	}

	if (!unlink($file)) {
		showErrorPage("$fileの削除に失敗しました。");
		exit(0);
	}

	printHeader("ラウンジ(削除)");
	print "<div class='header_links'>\n";
	print " <a href='wwwlng.cgi'>[戻る]</a>\n";
	print "</div>\n";
	print "$file の削除が完了しました。\n";
	printFooter();
}

#
# ファイルの更新時刻を合わせる
#
sub utimeTopic {
	my($passwd, @dirs) = @_;
	my($dir, $file, $date, $hflag, $mtime);
	local(*IN);

	# パスワードをチェックする
	if (($g_passwd eq "") || ($passwd ne $g_passwd)) {
		showErrorPage("パスワードが違います。");
		exit(0);
	}

	foreach $dir (@dirs) {
		if ($dir =~ /(^\/|\.\.)/) {
			showErrorPage("他のフォルダを参照することはできません。");
			exit(0);
		}
	}

	printHeader("ラウンジ(管理者用：ファイルの時刻調整)");
	print "<div class='header_links'>\n";
	print " <a href='wwwlng.cgi'>[戻る]</a>\n";
	print "</div>\n";

	if ($dirs[0] eq "all") {
		@dirs = ();
		if (!opendir(DIR, ".")) {
			showErrorPage("フォルダ情報の取得に失敗しました。");
			exit(0);
		}
		while ($dir = readdir(DIR)) {
			if ($dir =~ /^\./) { next; }
			if (! -d $dir) { next; }
			push(@dirs, $dir);
		}
		closedir(DIR);
	}

	foreach $dir (@dirs) {
		if (!opendir(DIR, $dir)) {
			next;
		}
		while ($file = readdir(DIR)) {
			if ($file =~ /^\./) { next; }
			if (! -f "$dir/$file") { next; }
			$file = "$dir/$file";
			$date = "";
			if (!open(IN, $file)) {
				next;
			}
			while (<IN>) {
				if (/^========================================[\r\n]+$/) {
					$hflag = 1;
				}
				if (/^[\r\n]+$/) {
					$hflag = 0;
				}
				if ($hflag == 0) {
					next;
				}
				if (/^Date: /) {
					$date = $_;
				}
			}
			close(IN);
			if ($date =~ /(\d\d\d\d)\/(\d\d)\/(\d\d).*(\d\d):(\d\d):(\d\d)/) {
				$mtime = mktime($1 - 1900, $2 - 1, $3, $4, $5, $6, -9);
				if (utime($mtime, $mtime, $file)) {
					print "$file $date OK.<br>\n";
				} else {
					print "$file $date NG.<br>\n";
				}
			}
		}
	}

	printFooter();
}

#
# トピック一覧を書き出す
#
sub printTopicList {
	my(@files) = @_;
	my($file, $hflag, $count, $questioner, $solved, $from, $date, $subject);
	my($line);

	print "<form method='POST' name='f1' action='wwwlng.cgi?search'>\n";
	print "<div class='search_block'>\n";
	print " <input type='text' size=30 name='word' placeholder='検索'> ";
	print " <button type='submit'>検索</button>\n";
	print "</div>\n";
	print "<table class='topic_list'>\n";
	print " <tr>\n";
	print "  <th>最終更新日</th>\n";
	print "  <th>解決</th>\n";
	if ($g_show_questioner) {
		print "  <th>質問者</th>\n";
	}
	print "  <th>最終発言者</th>\n";
	print "  <th>タイトル</th>\n";
	print " </tr>\n";
	$line = 0;
	foreach $file (@files) {
		$file = (split(/ /, $file))[1];
		open(IN, $file) || next;
		$hflag = 0;
		$count = 0;
		$questioner = "";
		$solved = "　";
		$from = "　";
		$date = "　";
		$subject = <IN>;
		$subject =~ s/^Subject: //;
		$subject =~ s/[\r\n]+$//;
		while (<IN>) {
			if (/^========================================[\r\n]+$/) {
				$hflag = 1;
			}
			if (/^[\r\n]+$/) {
				$hflag = 0;
			}
			if ($hflag == 0) {
				next;
			}
			if (/^From: (.*)[\r\n]+$/) {
				$from = $1;
				if (!$questioner) {
					$questioner = $from;
				}
				$count++;
			}
			if (/^Solved:/) {
				$solved = "済";
			}
			if (/^Date: (\d+)\/(\d+)\/(\d+)/) {
				$date = "$1/$2/$3";
			}
		}
		print " <tr>\n";
		print "  <td class='date'>$date</td>\n";
		print "  <td class='solved'>$solved</td>\n";
		if ($g_show_questioner) {
			print "  <td class='questioner'>$questioner</td>\n";
		}
		print "  <td class='last_updater'>" . html($from) . "</td>\n";
		print "  <td class='title'><a href='wwwlng.cgi?print+$file' title='$file'>" . html($subject) .  "</a> ($count件)</td>\n";
		print "</tr>\n";
		close(IN);
	}
	print "</table>\n";
	print "</form>\n";
}

#
# トピック内容を書き出す
#
sub printTopicContent {
	local(*IN) = @_;
	my($html, $topic_flag, $header_flag);

	$topic_flag = 0;
	$header_flag = 0;

	while (<IN>) {
		s/[\r\n]+//;
		if (/^Reply-To: /) {
			next;
		} elsif (/^From: /) {
			if ($topic_flag) {
				print " </div>\n";
				print "</div>\n";
				$topic_flag = 0;
			}
			s/^From: //;
			$topic_flag = 1;
			$header_flag = 1;
			print "<div class='topic'>\n";
			print " <div class='topic_header'>\n";
			print "  <span class='topic_name'>" . html($_) . "</span>\n";
		} elsif (/^E-Mail: /) {
			s/^E-Mail: //;
			print "  <a class='topic_email' href='mailto:" . html($_) . "'>[E-Mail]</a>\n";
		} elsif (/^Date: /) {
			s/^Date: //;
			print "  <span class='topic_date'>$_</span>\n";
		} elsif (/^Solved: /) {
			print "  <span class='topic_solved'>[[解決]]</span>\n";
		} elsif (/^ (.*)/) {
			$html = html($1);
			$html =~ s/(https?:\/\/[a-zA-Z0-9.\/\-+#_?~&%=^\@:;]+)/<a href="$1">$1<\/a>/g;
			if ($header_flag) {
				print " </div>\n";
				print " <div class='topic_body'>\n";
				$header_flag = 0;
			}
			print "$html<br>\n";
		}
	}
	if ($header_flag) {
		print " </div>\n";
	}
	if ($topic_flag) {
		print " </div>\n";
		print "</div>\n";
	}
}

#
# 検索結果を書き出す
#
sub printSearchResult {
	my($file, @words) = @_;
	my($subject, $wcount, $word, $text);
	my(%wordflag, %wordtext);

	$wcount = 0;
	open(IN, $file);
	$subject = <IN>;
	$subject =~ s/^Subject: //;
	$subject =~ s/[\r\n]+$//;
	seek(IN, 0, 0);
	loop: while (<IN>) {
		foreach $word (@words) {
			if ($wordflag{$word} == 1) { next; }
			if (!/$word/) { next; }
			s/[\r\n]+$//;
			$text = $_;
			$text =~ s/^(Subject|Date|From|E-Mail): //;
			$text =~ s/^ //;
			$wordtext{$word} = $text;
			$wordflag{$word} = 1;
			if (($FORM{'ANDOR'} eq "or") || (++$wcount == $#words + 1)) {
				if ($subject eq "") { $subject = $file; }
				print "<dt><a href='wwwlng.cgi?print+$file'>" . html($subject) . "</a>\n";
				foreach $word (@words) {
					print "<dd>" . html($wordtext{$word}) . "\n";
				}
				last loop;
			}
		}
	}
	close(IN);
}

#
# 新規ファイル名を決定する
#
sub determineFileName {
	my($sec, $min, $hour, $mday, $mon, $year);
	my($dir, $file, $num);

	# ディレクトリ名(YYYYMM)を得る
	$dir = formatDateTime("YYYYMM", time());

	# ディレクトリが無ければ作成する
	if (!-d $dir) {
		mkdir($dir, 0755);
	}

	# 最新のファイル番号を得る
	$num = 0;
	opendir(DIR, $dir);
	foreach $file (readdir(DIR)) {
		if ($file =~ /^[0-9]{4}([0-9]{4}).txt$/) {
			$num = ($num < $1) ? $1 : $num;
		}
	}
	closedir(DIR);
	$num++;

	# ファイル名(YYYYMM/YYMMNNNN.txt)を返す
	return(sprintf("$dir/%s%04d.txt", $dir, $num));
}

#
# ヘッダを書き出す
#
sub printHeader {
	my($title) = @_;
	print "Content-Type: text/html\n";
	print "\n";
	print "<!doctype html>\n";
	print "<html>\n";
	print "<head>\n";
	print "<meta charset='utf-8'>\n";
	print "<title>$title</title>\n";
	print "<meta name='viewport' content='width=device-width,initial-scale=1.0'>\n";
	print "<link rel='stylesheet' href='style.css'>\n";
	print "</head>\n";
	print "<body>\n";
	print "<h1>" . html($title) . "</h1>\n";
}

#
# フッタを書き出す
#
sub printFooter
{
	print "<hr>\n";
	print "<footer>\n";
	print "$g_name\n";
	print "</footer>\n";
	print "</body>\n";
	print "</html>\n";
}

#
# エラーページを表示する
#
sub showErrorPage {
	my($errmsg) = @_;
	printHeader("ラウンジ(エラー)");
	print "<div class='header_links'>\n";
	print " <a href='wwwlng.cgi'>[戻る]</a>\n";
	print "</div>\n";
	print "<div class='error_message'>$errmsg</div>\n";
	printFooter();
}

#
# フォームデータを読みこむ
#
sub readForm
{
	my($query_string, @a, $x, $name, $value);

	if ($ENV{'REQUEST_METHOD'} ne "POST") {
		return("");
	}
	if (!read(STDIN, $query_string, $ENV{'CONTENT_LENGTH'})) {
		print "111<br>\n"; exit(0);
		return("Can't read query string.");
	}
	@a = split(/&/, $query_string);
	foreach $x (@a) {
		($name, $value) = split(/=/, $x);
		$value =~ tr/+/ /;
		$value =~ s/%([0-9a-fA-F][0-9a-fA-F])/pack("C", hex($1))/eg;
		$value =~ s/\r\n/\n/g;
		$value =~ s/\s+$//;
		$FORM{$name} = $value;
	}
	return("");
}

#
# ファイルをロックする
#
sub fileLock
{
	for (my $i = 1; $i <= 6; $i++) {
		if (mkdir($g_lockfile, 0755)) {
			$SIG{'PIPE'} = "sigexit";
			$SIG{'INT'} = "sigexit";
			$SIG{'HUP'} = "sigexit";
			$SIG{'QUIT'} = "sigexit";
			$SIG{'TERM'} = "sigexit";
			last;
		} elsif ($i < 6) {
			sleep(2);
		} else {
			return(undef);
		}
	}
	return(1);
}

#
# ファイルのロックを解除する
#
sub fileUnlock
{
	for (my $i = 1; $i <= 4; $i++) {
		if (rmdir($g_lockfile) == 0) {
			last;
		} elsif ($i < 4) {
			sleep(1);
		} else {
			return(undef);
		}
	}
	return(1);
}

#
# プログラム中断時の処理
#
sub sigexit
{
	if (-d $g_lockfile) {
		rmdir($g_lockfile);
	}
	exit(0);
}

#
# HTMLサニタイジング・
#
sub html {
    my($msg) = $_[0];
	$msg =~ s/&/&amp;/g;
	$msg =~ s/</&lt;/g;
	$msg =~ s/>/&gt;/g;
	$msg =~ s/"/&quot;/g;
	$msg =~ s/'/&#39;/g;
	$msg =~ s/ /&nbsp;/g;
	return $msg;
}
#
# 年/月/日 時:分:秒から、UNIX時間(1970/1/1からの秒数)を求める
#
sub mktime {
	my($year, $mon, $mday, $hour, $min, $sec, $tz) = @_;
	my(@mdays, $xtime, $uru);

	@mdays = ( 0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334);
	$uru = int(($year - 72) / 4);
	if (((($year + 1900) % 4) == 0) && ($mon < 2)) { $uru--; }
	$xtime = $year - 70;
	$xtime = ($xtime * 365) + $mdays[$mon] + $mday + $uru;
	$xtime = ($xtime * 24) + $hour + $tz;
	$xtime = ($xtime * 60) + $min;
	$xtime = ($xtime * 60) + $sec;
	return($xtime);
}

#
# 日時をフォーマットする
#
sub formatDateTime {
	my($format, $mtime) = @_;
	my($sec, $min, $hour, $mday, $mon, $year, $wday) = localtime($mtime);
	if ($format eq "YYYYMMDDhhmmss") {
		return sprintf("%04d%02d%02d%02d%02d%02d",
				$year + 1900, $mon + 1, $mday, $hour, $min, $sec);
	} elsif ($format eq "YYYYMM") {
		return sprintf("%04d%02d", $year + 1900, $mon + 1);
	} elsif ($format eq "YYYY/MM/DD(A) hh:mm:ss") {
		return sprintf("%04d/%02d/%02d(%s) %02d:%02d:%02d",
				$year + 1900, $mon + 1, $mday, $g_wdays[$wday], $hour, $min, $sec);
	} else {
		showErrorPage("時刻フォーマットが不正です。");
		exit(0);
	}
}

